from abc import ABC
from typing import List

from palaestrai.agent import RewardInformation, SensorInformation


class Reward(ABC):
    """Abstract base class for environment rewards.

    The :class:`Reward` ABC is one way to implement a new environment
    reward. The reward is calculated from the environment view and
    measure the environment performance only. The agents use this
    reward to calculate their own internal reward by using the
    objective.

    To use the reward, it has to be called with a list of
    :class:`SensorInformation` objects::

        reward = MyReward()
        value = reward([SensorInformation(...), SensorInformation(...)])

    Parameters
    ----------
    params: dict
        A dict that can be used to pass configuration parameters to
        the reward instance.

    """

    def __init__(self, **kwargs):
        self.params: dict = kwargs

    def __call__(
        self, state: List[SensorInformation], *args, **kwargs
    ) -> List[RewardInformation]:
        raise NotImplementedError



"""This module contains exceptions."""


class ShutdownError(Exception):
    """This error is raised to kill mosaik on purpose.

    Mosaik will not see the error raising and simply crash (in the same
    way mosaik is killed with a keyboard interrupt).

    No responsibility is taken for any damage on unfinished simulators
    waiting for their next step.

    Such a sadness!

    """



"""This module contains a modified version of the mosaik scheduler.

Since the scheduler is not a class but a collection of functions,
each function of the scheduler needs to be present. For most of them,
calls can be directed to the regular scheduler function. However, as
soon as the mosaik scheduler changes, these changes need to be
reflected here and new functions need a forward call in this module.

The only function with a modification is the
:func:`mosaik.scheduler.get_input_data` function (and there is added
only one line of code). But to make use of this modified function, it
is required to provide copies of :func:`mosaik.scheduler.run` and
:func:`mosaik.scheduler.sim_process`, as well. They should be updated
on a new mosaik version.

See
https://mosaik.readthedocs.io/en/latest/api_reference/mosaik.scheduler.html
for the docs of the original functions.

Tested with mosaik version 2.6.0

"""
from time import perf_counter

import mosaik.scheduler
import pkg_resources

# Get major mosaik version
MM_VERSION = int(pkg_resources.require("mosaik")[0].version.split(".")[0])
if MM_VERSION >= 3:
    from mosaik.exceptions import (
        NoStepException,
        SimulationError,
        WakeUpException,
    )

from palaestrai_mosaik.mosaikpatch import LOG
from simpy.exceptions import Interrupt


def run(
    world,
    until,
    rt_factor=None,
    rt_strict=False,
    print_progress=True,
    lazy_stepping=True,
):
    """Run the simulation.

    This is an exact copy of the current :func:`mosaik.scheduler.run`
    method.

    """

    if MM_VERSION < 3:
        print(
            "It seems you're not using the latest Mosaik version. "
            "Please consider updating:\n\n\tpip install --upgrade mosaik\n"
        )
    world.until = until
    if rt_factor is not None and rt_factor <= 0:
        raise ValueError('"rt_factor" is %s but must be > 0"' % rt_factor)
    if rt_factor is not None:
        # Adjust rt_factor to the time_resolution:
        rt_factor *= world.time_resolution
    world.rt_factor = rt_factor

    env = world.env

    setup_done_events = []
    for sim in world.sims.values():
        if sim.meta["api_version"] >= (2, 2):
            # setup_done() was added in API version 2.2:
            setup_done_events.append(sim.proxy.setup_done())

    yield env.all_of(setup_done_events)

    processes = []
    for sim in world.sims.values():
        if MM_VERSION >= 3:
            process = env.process(
                sim_process3(
                    world,
                    sim,
                    until,
                    rt_factor,
                    rt_strict,
                    print_progress,
                    lazy_stepping,
                )
            )
        else:
            process = env.process(
                sim_process2(
                    world, sim, until, rt_factor, rt_strict, print_progress
                )
            )

        sim.sim_proc = process
        processes.append(process)

    yield env.all_of(processes)


def sim_process2(world, sim, until, rt_factor, rt_strict, print_progress):
    """SimPy simulation process for a certain simulator *sim*.

    This method is an exact copy of the current
    :func:`mosaik.scheduler.sim_process` method.

    """

    rt_start = perf_counter()

    try:
        keep_running = get_keep_running_func(world, sim, until)
        while keep_running():
            try:
                yield step_required(world, sim)
            except StopIteration:
                # We've been woken up by a terminating successor.
                # Check if we can also stop or need to keep running.
                continue

            yield wait_for_dependencies(world, sim)
            input_data = get_input_data(world, sim)
            yield from rt_sleep(rt_factor, rt_start, sim, world)
            yield from step(world, sim, input_data)
            rt_check(rt_factor, rt_start, rt_strict, sim)
            yield from get_outputs(world, sim)
            world.sim_progress = get_progress(world.sims, until)
            if print_progress:
                print("Progress: %.2f%%" % world.sim_progress, end="\r")

        # Before we stop, we wake up all dependencies who may be waiting for
        # us. They can then decide whether to also stop of if there's another
        # process left for which they need to provide data.
        for pre_sid in world.df_graph.predecessors(sim.sid):
            evt = world.sims[pre_sid].step_required
            if not evt.triggered:
                evt.fail(StopIteration())

    except ConnectionError as err:
        raise SimulationError(
            f"Simulator '{sim.sid}' closed its connection.", err
        )


def sim_process3(
    world, sim, until, rt_factor, rt_strict, print_progress, lazy_stepping
):
    """
    SimPy simulation process for a certain simulator *sim*.
    """
    sim.rt_start = rt_start = perf_counter()

    try:
        keep_running = get_keep_running_func(
            world, sim, until, rt_factor, rt_start
        )
        while keep_running():
            try:
                yield from has_next_step(world, sim)
            except WakeUpException:
                # We've been woken up by a terminating predecessor.
                # Check if we can also stop or need to keep running.
                continue
            except NoStepException:
                # None of the simulators has a next step, therefore stop.
                break
            sim.interruptable = True
            while True:
                try:
                    yield from rt_sleep(rt_factor, rt_start, sim, world)
                    yield wait_for_dependencies(world, sim, lazy_stepping)
                    break
                except Interrupt as i:
                    assert i.cause == "Earlier step"
                    clear_wait_events(sim)
                    continue
            sim.interruptable = False
            input_data = get_input_data(world, sim)
            max_advance = get_max_advance(world, sim, until)
            yield from step(world, sim, input_data, max_advance)
            rt_check(rt_factor, rt_start, rt_strict, sim)
            yield from get_outputs(world, sim)
            notify_dependencies(world, sim)
            if world._df_cache:
                prune_dataflow_cache(world)
            world.sim_progress = get_progress(world.sims, until)
            if print_progress:
                print("Progress: %.2f%%" % world.sim_progress, end="\r")
        sim.progress_tmp = until
        sim.progress = until
        clear_wait_events_dependencies(sim)
        check_and_resolve_deadlocks(sim, end=True)
        # Before we stop, we wake up all dependencies who may be waiting for
        # us. They can then decide whether to also stop of if there's another
        # process left which might provide data.
        for suc_sid in world.trigger_graph.successors(sim.sid):
            if not world.sims[suc_sid].sim_proc.triggered:
                evt = world.sims[suc_sid].has_next_step
                if not evt.triggered:
                    world.sims[suc_sid].sim_proc.interrupt("Stopped simulator")

    except ConnectionError as e:
        raise SimulationError(
            'Simulator "%s" closed its connection.' % sim.sid, e
        )


def get_input_data(world, sim):
    """Return a dictionary with the input data for *sim*.

    The original function :func:`mosaik.scheduler.get_input_data` is
    called. Afterwards, the :func:`~.trigger_actuators` method of the
    modified world object is called.

    """
    input_data = mosaik.scheduler.get_input_data(world, sim)

    LOG.debug("Now triggering actuators for sim %s", sim.sid)

    # This is the modification for ARL
    world.trigger_actuators(sim, input_data)
    # This was the modification for ARL

    return input_data


def get_next_step(sim):
    return mosaik.scheduler.get_next_step(sim)


def has_next_step(world, sim):
    return mosaik.scheduler.has_next_step(world, sim)


def get_max_advance(world, sim, until):
    return mosaik.scheduler.get_max_advance(world, sim, until)


def treat_cycling_output(world, sim, data, output_time):
    return mosaik.scheduler.treat_cycling_output(world, sim, data, output_time)


def notify_dependencies(world, sim):
    return mosaik.scheduler.notify_dependencies(world, sim)


def prune_dataflow_cache(world):
    return mosaik.scheduler.prune_dataflow_cache(world)


def check_and_resolve_deadlocks(sim, waiting=False, end=False):
    return mosaik.scheduler.check_and_resolve_deadlocks(sim, waiting, end)


def clear_wait_events(sim):
    return mosaik.scheduler.clear_wait_events(sim)


def clear_wait_events_dependencies(sim):
    return mosaik.scheduler.clear_wait_events_dependencies(sim)


def step(world, sim, inputs, max_advance=0):
    """Step the scheduler.

    See :func:`mosaik.scheduler.step`.

    """
    if MM_VERSION >= 3:
        return mosaik.scheduler.step(world, sim, inputs, max_advance)
    else:
        return mosaik.scheduler.step(world, sim, inputs)


def get_outputs(world, sim):
    """Get outputs for a simulator.

    See :func:`mosaik.scheduler.get_outputs`.

    """
    return mosaik.scheduler.get_outputs(world, sim)


def get_progress(sims, until):
    """Get simulation progress.

    See :func:`mosaik.scheduler.get_progress`.

    """
    return mosaik.scheduler.get_progress(sims, until)


def rt_sleep(rt_factor, rt_start, sim, world):
    """Sleep for real time simulation.

    See :func:`mosaik.scheduler.rt_sleep`.

    """
    return mosaik.scheduler.rt_sleep(rt_factor, rt_start, sim, world)


def rt_check(rt_factor, rt_start, rt_strict, sim):
    """Check the time for real time simulation.

    See :func:`mosaik.scheduler.rt_check`.

    """
    return mosaik.scheduler.rt_check(rt_factor, rt_start, rt_strict, sim)


def get_keep_running_func(world, sim, until, rt_factor=1.0, rt_start=0):
    """Return the *keep_running_func*.

    See :func:`mosaik.scheduler.get_keep_running_func`.

    """
    if MM_VERSION >= 3:
        return mosaik.scheduler.get_keep_running_func(
            world, sim, until, rt_factor, rt_start
        )
    else:
        return mosaik.scheduler.get_keep_running_func(world, sim, until)


def step_required(world, sim):
    """Return if another step is required.

    See :func:`mosaik.scheduler.step_required`.

    """
    return mosaik.scheduler.step_required(world, sim)


def wait_for_dependencies(world, sim, lazy_stepping=True):
    """Wait for dependencies of *sim*.

    See :func:`mosaik.scheduler.wait_for_dependencies`.

    """
    if MM_VERSION >= 3:

        return mosaik.scheduler.wait_for_dependencies(
            world, sim, lazy_stepping
        )
    else:
        return mosaik.scheduler.wait_for_dependencies(world, sim)



"""This module contains the debug script that can be used to start
the dummy experiment without the CLI.

"""
import asyncio
import logging
import os
import sys

from palaestrai.cli.manager import init_logger
from palaestrai.core import RuntimeConfig
from palaestrai.experiment.executor import Executor, ExecutorState, Experiment


def main(filename="dummy_experiment", verbose=0, use_custom_logging=False):
    # runtime_file = os.path.abspath(
    #     os.path.join(
    #         __file__, "..", "..", "fixtures", "arl-runtime-debug.conf.yaml"
    #     )
    # )
    # with open(runtime_file) as f:
    #     conf = f.read()

    RuntimeConfig().load()

    init_logger(verbose)
    if use_custom_logging:
        custom_logging()

    # experiment_file = os.path.abspath(
    #     os.path.join(__file__, "..", "..", "fixtures", f"{filename}.yml")
    # )
    try:
        experiment = Experiment.load(filename)
    except (NotADirectoryError, FileNotFoundError):
        experiment_file = os.path.abspath(
            os.path.join(__file__, "..", "..", "fixtures", f"{filename}.yml")
        )
        experiment = Experiment.load(experiment_file)
        print(
            "Ignore that stupid, annoying error message. "
            "I already found the file!!1."
        )

    executor = Executor()
    executor.schedule(experiment)
    executor_final_state = asyncio.run(executor.execute())
    if executor_final_state != ExecutorState.EXITED:
        sys.exit(
            {
                ExecutorState.SIGINT: -2,
                ExecutorState.SIGABRT: -6,
                ExecutorState.SIGTERM: -15,
            }[executor_final_state]
        )


def custom_logging():
    """Manually configure palaestrai loggers."""
    logging.getLogger("palaestrai").setLevel("ERROR")
    logging.getLogger("palaestrai.agent").setLevel("ERROR")
    logging.getLogger("palaestrai.core").setLevel("ERROR")
    logging.getLogger("palaestrai.store").setLevel("ERROR")
    logging.getLogger("palaestrai.experiment").setLevel("ERROR")
    logging.getLogger("palaestrai.environment").setLevel("ERROR")
    logging.getLogger("palaestrai.util").setLevel("ERROR")
    logging.getLogger("palaestrai_mosaik").setLevel("DEBUG")
    logging.getLogger("palaestrai_mosaik.environment").setLevel("ERROR")
    logging.getLogger("palaestrai_mosaik.mosaikpatch").setLevel("ERROR")
    logging.getLogger("palaestrai_mosaik.simulator").setLevel("ERROR")

    logging.getLogger("pandapower").setLevel("ERROR")
    logging.getLogger("matplotlib").setLevel("ERROR")
    logging.getLogger("h5py").setLevel("ERROR")
    logging.getLogger("midas").setLevel("DEBUG")
    logging.getLogger("pysimmods").setLevel("ERROR")
    logging.getLogger("numexpr").setLevel("ERROR")


if __name__ == "__main__":
    # main("dummy_experiment", use_custom_logging=True)
    main("example_experiment_midas", use_custom_logging=True)



"""This module contains a patch for mosaik's world object.

Needs to be updated when the mosaik API changes.

Tested with mosaik version 2.6.0.

"""
import types

import networkx
import pkg_resources
from mosaik import util
from palaestrai_mosaik.mosaikpatch import LOG, scheduler


def modify_world(world):
    """Modify the world object for the use with palaestrAI.

    A new method :meth:`.trigger_actuators` is installed, which is
    called by the :func:`.get_input_data` function of the modified
    :mod:`~.scheduler`. Furthermore, the
    :func:`mosaik.scenario.World.run` method is overwritten by the
    modified :func:`~.run` version below.

    Parameters
    ----------
    world : :class:`mosaik.scenario.World`
        The *world* object to be modified.

    """
    LOG.debug("Installing 'trigger_actuators' method.")
    setattr(world, "trigger_actuators", None)
    LOG.debug("Modifying 'run' method.")
    setattr(world, "run", types.MethodType(run, world))


def run(
    self,
    until,
    rt_factor=None,
    rt_strict=False,
    print_progress=True,
    lazy_stepping=True,
):
    """A modified version of the :func:`mosaik.scenario.World.run`
    method of mosaik's world object.

    This :func:`~.run` method works exactly like the original one
    except that a different scheduler is used. Indeed, it is nearly the
    same code but since a modified scheduler is imported, this specific
    scheduler is called instead.

    Furthermore, the prints are replaced by logging outputs.

    See the mosaik documentation
    https://mosaik.readthedocs.io/en/latest/api_reference/mosaik.scenario.html
    :func:`mosaik.scenario.World.run` for the regular features of this
    method.

    """
    if self.srv_sock is None:
        raise RuntimeError(
            "Simulation has already been run and can only "
            "be run once for a World instance."
        )

    # Check if a simulator is not connected to anything:
    for sid, deg in sorted(list(networkx.degree(self.df_graph))):
        if deg == 0:
            LOG.warning("WARNING: %s has no connections.", sid)

    # Get major mosaik version
    mm_version = int(pkg_resources.require("mosaik")[0].version.split(".")[0])
    if mm_version >= 3:
        self.detect_unresolved_cycles()

        trigger_edges = [
            (u, v)
            for (u, v, w) in self.df_graph.edges.data(True)
            if w["trigger"]
        ]
        self.trigger_graph.add_edges_from(trigger_edges)

        self.cache_trigger_cycles()
        self.cache_dependencies()
        self.cache_related_sims()
        self.cache_triggering_ancestors()
        self.create_simulator_ranking()

    LOG.debug("Starting simulation.")
    import mosaik._debug as dbg  # always import, enable when requested

    if self._debug:
        dbg.enable()
    try:
        if mm_version >= 3:
            util.sync_process(
                scheduler.run(
                    self,
                    until,
                    rt_factor,
                    rt_strict,
                    print_progress,
                    lazy_stepping,
                ),
                self,
            )
        else:
            util.sync_process(
                scheduler.run(
                    self, until, rt_factor, rt_strict, print_progress
                ),
                self,
            )
        LOG.debug("Simulation finished successfully.")
    except KeyboardInterrupt:
        LOG.info("Simulation canceled. Terminating ...")
    finally:
        self.shutdown()
        if self._debug:
            dbg.disable()



"""This module contains the function :func:`.load_funcs` to load a
mosaik environment.

This comprises to instantiate a world object und retrieve sensor and
actuator descriptions

"""
from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

if TYPE_CHECKING:
    from mosaik import World
    from palaestrai.agent.actuator_information import ActuatorInformation
    from palaestrai.agent.sensor_information import SensorInformation


def load_funcs(
    module_name: str, description_func: str, instance_func: str
) -> Tuple[
    Callable[[Dict[str, Any]], World],
    Callable[
        [Dict[str, Any]],
        Tuple[
            List[Union[SensorInformation, Dict[str, Any]]],
            List[Union[ActuatorInformation, Dict[str, Any]]],
        ],
    ],
]:
    """Load the description functions.

    Expects a dictionary containing the keys *"module"*,
    *"description_func"*, and "instance_func". *"module"* can
    either be a python module or a python class. The path segments
    for modules are separated by a dot "." and a class is separated
    by a colon ":", e.g., if *descriptor* is a module::

        {
            "module": "midas.adapter.harlequin.descriptor",
            "description_func": "describe",
            "instance_func": "get_world",
        }

    or, if *Descriptor* is a class::

        {
            "module": "midas.adapter.harlequin:Descriptor",
            "description_func": "describe",
            "instance_func": "get_world",
        }


    Parameters
    ----------
    params : dict
        A *dict* containing the keys as described above.

    Returns
    -------
    tuple
        A *tuple* of the description function and the instance
        function.

    """

    if ":" in module_name:
        module, clazz = module_name.split(":")
        module = import_module(module)
        obj = getattr(module, clazz)()
    else:
        obj = import_module(module_name)

    dscr_func = getattr(obj, description_func)
    inst_func = getattr(obj, instance_func)

    return dscr_func, inst_func



"""This module contains the :class:`.MosaikWorld` class, which is home
of the mosaik world object and cares for setup and shutdown of the
world.

"""
from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import numpy
from mosaik.exceptions import ScenarioError
from mosaik.scenario import Entity

from ..config import AGENT_SID
from ..mosaikpatch.scenario import modify_world
from ..simulator import ShutdownError
from . import LOG, loader
from .mosaik_ifs import MosaikInterfaceComponent

if TYPE_CHECKING:
    import mosaik

    from ..mosaik_environment import MosaikEnvironment


class MosaikWorld:
    """The home for mosaik worlds.

    This class creates sensor and actuator descriptions and the world
    objectby calling the :func:`description_func` and
    :func:`instance_func` defined in the params dictionary.

    Furthermore, the :class:`.ARLSyncSimulator` is injected into the
    world object and integrated into the mosaik dependency graph.

    Parameters
    ----------
    env : :class:`.MosaikEnvironment`
        A reference to the environment creating this object.

    Attributes
    ----------
    env : :class:`.MosaikEnvironment`
        Stores the reference to the environment creating this object.
    world : :class:`mosaik.scenario.World`
        The world object created during the setup method.
    end : int
        The number of simulation steps mosaik will simulate.
    mifs : :class:`.MosaikInterfaceComponent`
        Stores all sensors and actuators and manages their data.

    """

    def __init__(self, env):
        self.env: MosaikEnvironment = env
        self.world: mosaik.World
        self.end: int
        self.mifs: MosaikInterfaceComponent
        self.now_dt: datetime
        self.simtime: int = 0

    def setup(
        self,
        module_name: str,
        description_func: str,
        instance_func: str,
        end: Union[str, int],
        step_size: int,
        mosaik_params: Dict[str, Any],
    ):
        """Setup the mosaik environment.

        A couple of things need to be done to prepare the environment.

        Parameters
        ----------
        params : dict
            A *dict* with information about the mosaik world to be
            created.

        """
        self.end = get_end(end)
        self.now_dt = get_start(mosaik_params, self.env.rng)

        dscr_func, inst_func = loader.load_funcs(
            module_name, description_func, instance_func
        )
        sen_dscr, act_dscr = dscr_func(mosaik_params)

        self.world = inst_func(mosaik_params)
        modify_world(self.world)

        self.mifs = MosaikInterfaceComponent(sen_dscr, act_dscr)
        self.mifs.create_sensors()
        self.mifs.create_actuators()

        agent = self.inject_arl(step_size)
        self.connect(agent)

        self.world.trigger_actuators = self.mifs.trigger_actuators

    def inject_arl(self, step_size):
        """Prepare the world for ARL.

        The world object is extended by an instance of the
        :class:`.ARLSyncSimulator` which controls an *ARLAgent* mosaik
        entity.

        Returns
        -------
        :class:`mosaik.scenario.Entity`
            The agent entity object created by mosaik.

        """

        self.world.sim_config[AGENT_SID] = {
            "python": "palaestrai_mosaik.simulator:ARLSyncSimulator"
        }
        arl_sim = self.world.start(
            AGENT_SID, env=self.env, step_size=step_size
        )

        return arl_sim.ARLAgent()

    def connect(self, agent):
        """Establishs mosaik connections from sensor entities.

        The collected data is not used since the data is directly
        gathered from the world object. Instead, this step is required
        to get the simulator in sync with the other simulators. For
        this purpose, it would be enought to connect to a single other
        simulator, but to be sure we connect to every simulator that
        provides sensors.

        This is also on reason, why there will be no actuator
        connections. Furthermore, this would require to make use of
        the :attr:`time_shifted` parameter of the
        :meth:`mosaik.scenario.World.connect` method, which needs to be
        provided with :attr:`initial_data`, which we not have.
        Eventually, the actuator data will be injected directly into
        the world object.

        Notes
        -----
            While the *sid* has kind of standardized naming, this is
            not true for the eid, since it is task of each simulator to
            provide eids. Therefore, this might fail, although the
            sensor description SHOULD have the correct name.

            The *connect* method of the mosaik world currently does
            not use the entity's children. Therefore, passing an empty
            list is not a problem. This may change in future mosaik
            versions.

        Parameters
        ----------
        agent : :class:`mosaik.scenario.Entity`
            The agent entity to be connected with other entities.

        """
        connections = 0

        for key in self.mifs.sensor_map:
            sid, eid, attr = key.split(".")
            sim = self.world.sims[sid]

            model_type = eid
            for model in sim.meta["models"]:
                if model.lower() in eid.lower():
                    model_type = model

            entity = Entity(
                sid=sid,
                eid=eid,
                sim_name=sid.split("-")[0],
                type=model_type,
                children=list(),
                sim=sim,
            )
            try:
                self.world.connect(entity, agent, attr)
                connections += 1
            except ScenarioError as err:
                LOG.debug("Can't find model for eid %s", eid)
                LOG.debug("Caught the following exception: %s", err)
        if connections < 1:
            LOG.error(
                "Could not create any connection to any mosaik simulator. "
                "Please check your configuration and the log file."
            )
        else:
            LOG.debug(
                "Created %d connections to other simulators", connections
            )

    def start(self):
        """Start the mosaik simulation process."""
        LOG.debug("Starting mosaik ...")
        try:
            self.world.run(until=self.end)
            LOG.debug("Mosaik finished.")
        except ShutdownError:
            LOG.info("Successfully killed mosaik.")

    def update_sim_time(self, time):
        """Update the internal simtime.

        If now_dt is not None, it will be updated as well.
        """
        dif = max(0, time - self.simtime)
        self.simtime = time
        if self.now_dt is not None:
            self.now_dt += timedelta(seconds=dif)


def get_end(end: Union[str, int]):
    """Read the *end* value from the params dict.

    The *end* value is an integer, but sometimes it is provided
    as float, or as str like '15*60'. In the latter case, the
    str is evaluated (i.e., multiplied). In any case, *end* is
    returned as int.

    """
    if isinstance(end, str):
        parts = end.split("*")
        end = 1
        for part in parts:
            end *= float(part)
    return int(end)


def get_start(
    params: Dict[str, Any], rng: numpy.random.RandomState
) -> Optional[datetime]:
    """Search through the parameter to find a start_date.

    Start date has to be provided as ISO datestring like::

        2021-10-26 13:25:16+0100

    """
    if "start_date" in params:
        LOG.info("Found start_date %s", params["start_date"])
        if params["start_date"] == "random":
            params["start_date"] = (
                f"2020-{rng.randint(1, 12):02d}-"
                f"{rng.randint(1, 28):02d} "
                f"{rng.randint(0, 23):02d}:00:00+0100"
            )
            # params["start_date"] = self.base.start_date

        return datetime.strptime(params["start_date"], "%Y-%m-%d %H:%M:%S%z")

    for key, value in params.items():
        if isinstance(value, dict):
            res = get_start(value, rng)

            if res is not None:
                return res

    LOG.warning("Did not find start_date: %s", params)
    return None



"""This module contains the :class:`.ARLSyncSimulator`, which is used
by the :class:`.MosaikEnvironment` for synchronization.

"""
import mosaik_api
from palaestrai_mosaik.config import AGENT_EID
from palaestrai_mosaik.simulator import LOG

META = {
    "type": "time-based",
    "models": {
        AGENT_EID: {
            "public": True,
            "any_inputs": True,
            "params": ["env"],
            "attrs": [],
        }
    },
}


class ARLSyncSimulator(mosaik_api.Simulator):
    """A simulator for the synchronization of palaestrAI and mosaik.

    Attributes
    ----------
    sid : str
        The simulator id for this simulator given by mosaik
    step_size : int
        The step_size of this simulator
    models : dict
        A dictionary containing all models of this simulator.
        Currently, there is no reason why there should be more than one
        agent model.

    """

    def __init__(self):
        super().__init__(META)
        self.sid = None
        self.step_size = None
        self.models = dict()

        self._env = None

    def init(self, sid, **sim_params):
        """Initialize this simulator.

        Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator id provided by mosaik.

        Returns
        -------
        dict
            The meta description for this simulator as *dict*.

        """
        self.sid = sid
        self._env = sim_params["env"]
        self.step_size = sim_params["step_size"]

        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity)

        Parameters
        ----------
        num : int
            The number of models to create in one go.
        model : str
            The model to create. Needs to be present in the META.

        Returns
        -------
        list
            A *list* of the entities created during this call.

        """
        assert model == AGENT_EID

        entities = list()

        for _ in range(num):
            eid = f"{AGENT_EID}-{len(self.models)}"
            self.models[eid] = {
                "model": model,
                "eid": eid,
            }
            entities.append({"eid": eid, "type": model})
        return entities

    def step(self, time, inputs, max_advance=0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation time (the current step).
        inputs : dict
            A *dict* with inputs for the models.

        Returns
        -------
        int
            The simulation time at which this simulator should
            perform its next step.

        """

        LOG.debug("Stepped ARLSyncSim at step %d", time)

        self._env.sync(time)

        return time + self.step_size

    def get_data(self, outputs):
        """Return requested outputs (if feasible).

        Since this simulator does not generate output for its own, an
        empty dict is returned.

        Parameters
        ----------
        outputs : dict
            Requested outputs.

        Returns
        -------
        dict
            An empty dictionary, since no output is generated.

        """
        return dict()



import unittest
from unittest.mock import MagicMock, patch

from palaestrai_mosaik.mosaik_environment import MosaikEnvironment


class TestMosaikEnvironment(unittest.TestCase):
    @unittest.skip
    @patch(f"{MosaikEnvironment.__module__}.threading.Thread")
    @patch(f"{MosaikEnvironment.__module__}.MosaikWorld")
    def test_start(self, mock_world, mock_thread):
        env = MosaikEnvironment(
            "", "", 0, params={"reward": {"name": "", "params": dict()}}
        )
        mock_world.setup = MagicMock()

        env.start_environment()


if __name__ == "__main__":
    unittest.main()
